import sys
from pysnmp.smi.macro import rfc2579
from pysnmp.smi import error
from pysnmp import debug

OctetString, Integer, ObjectIdentifier = mibBuilder.importSymbols(
    'ASN1', 'OctetString', 'Integer', 'ObjectIdentifier'
    )
( NamedValues, ) = mibBuilder.importSymbols("ASN1-ENUMERATION", "NamedValues")
( ConstraintsIntersection, ConstraintsUnion, SingleValueConstraint,
  ValueRangeConstraint, ValueSizeConstraint, ) = mibBuilder.importSymbols(
    "ASN1-REFINEMENT", "ConstraintsIntersection", "ConstraintsUnion",
    "SingleValueConstraint", "ValueRangeConstraint", "ValueSizeConstraint"
    )
Counter32, Unsigned32, TimeTicks, Counter64 = mibBuilder.importSymbols(
    'SNMPv2-SMI', 'Counter32', 'Unsigned32', 'TimeTicks', 'Counter64'
    )

class DisplayString(rfc2579.TextualConvention, OctetString):
    subtypeSpec = OctetString.subtypeSpec+ValueSizeConstraint(0,255)
    displayHint = "255a"

class PhysAddress(rfc2579.TextualConvention, OctetString):
    displayHint = "1x:"

class MacAddress(rfc2579.TextualConvention, OctetString):
    subtypeSpec = OctetString.subtypeSpec+ValueSizeConstraint(6,6)
    displayHint = "1x:"
    fixedLength = 6

class TruthValue(Integer, rfc2579.TextualConvention):
    subtypeSpec = Integer.subtypeSpec+SingleValueConstraint(1, 2)
    namedValues = NamedValues(('true', 1), ('false', 2))
    
class TestAndIncr(Integer, rfc2579.TextualConvention):
    subtypeSpec = Integer.subtypeSpec+ValueRangeConstraint(0, 2147483647)
    defaultValue = 0
    def smiWrite(self, name, value, idx):
        if value != self:
            raise error.InconsistentValueError(idx=idx, name=name)
        value = value + 1
        if value > 2147483646:
            value = 0
        return self.clone(value)

class AutonomousType(ObjectIdentifier, rfc2579.TextualConvention): pass
class InstancePointer(ObjectIdentifier, rfc2579.TextualConvention):
    status = 'obsolete'
class VariablePointer(ObjectIdentifier, rfc2579.TextualConvention): pass
class RowPointer(ObjectIdentifier, rfc2579.TextualConvention): pass
    
class RowStatus(Integer, rfc2579.TextualConvention):
    """A special kind of scalar MIB variable responsible for
       MIB table row creation/destruction.
    """
    subtypeSpec = Integer.subtypeSpec+SingleValueConstraint(0, 1, 2, 3, 4, 5, 6)
    namedValues = NamedValues(
        ('notExists', 0), ('active', 1), ('notInService', 2), ('notReady', 3),
        ('createAndGo', 4), ('createAndWait', 5), ('destroy', 6)
        )
    # Known row states
    stNotExists, stActive, stNotInService, stNotReady, \
                 stCreateAndGo, stCreateAndWait, stDestroy = list(range(7))
    # States transition matrix (see RFC-1903)
    stateMatrix = {
        # (new-state, current-state)  ->  (error, new-state)
        ( stCreateAndGo, stNotExists ): (
        error.RowCreationWanted, stActive
        ),
        ( stCreateAndGo, stNotReady ): (
        error.InconsistentValueError, stNotReady
        ),
        ( stCreateAndGo, stNotInService ): (
        error.InconsistentValueError, stNotInService
        ),
        ( stCreateAndGo, stActive ): (
        error.InconsistentValueError, stActive
        ),
        #
        ( stCreateAndWait, stNotExists ): (
        error.RowCreationWanted, stActive
        ),
        ( stCreateAndWait, stNotReady ): (
        error.InconsistentValueError, stNotReady
        ),
        ( stCreateAndWait, stNotInService ): (
        error.InconsistentValueError, stNotInService
        ),
        ( stCreateAndWait, stActive ): (
        error.InconsistentValueError, stActive
        ),
        #
        ( stActive, stNotExists ): (
        error.InconsistentValueError, stNotExists
        ),
        ( stActive, stNotReady ): (
        error.InconsistentValueError, stNotReady
        ),
        ( stActive, stNotInService ): (
        None, stActive
        ),
        ( stActive, stActive ): (
        None, stActive
        ),
        #
        ( stNotInService, stNotExists ): (
        error.InconsistentValueError, stNotExists
        ),
        ( stNotInService, stNotReady ): (
        error.InconsistentValueError, stNotReady
        ),
        ( stNotInService, stNotInService ): (
        None, stNotInService
        ),
        ( stNotInService, stActive ): (
        None, stActive
        ),
        #
        ( stDestroy, stNotExists ): (
        error.RowDestructionWanted, stNotExists
        ),
        ( stDestroy, stNotReady ): (
        error.RowDestructionWanted, stNotExists
        ),
        ( stDestroy, stNotInService ): (
        error.RowDestructionWanted, stNotExists
        ),
        ( stDestroy, stActive ): (
        error.RowDestructionWanted, stNotExists
        ),
        # This is used on instantiation
        ( stNotExists, stNotExists ): (
        None, stNotExists
        )
        }
    defaultValue = stNotExists
    pendingError = None
    
    def smiWrite(self, name, value, idx):
        # Run through states transition matrix, resolve new instance value
        err, val = self.stateMatrix.get(
            (self.clone(value), int(self)), (error.MibOperationError, None)
            )
        debug.logger & debug.flagIns and debug.logger('RowStatus state resolution: %s, %s -> %s, %s' % (value, int(self), err, val))
        if val is None:
            val = self
        else:
            val = self.clone(val)
        if err is not None:
            err = err(
                msg='Exception at row state transition %s->%s' % (self, value),
                idx=idx
                )
            val.smiSetPendingError(err)
        return val

    def smiCreate(self, name, value, idx):
        return self.smiWrite(name, value, idx)
        
    def smiRaisePendingError(self):
        if self.pendingError:
            err, self.pendingError = self.pendingError, None
            raise err
    def smiSetPendingError(self, err):
        self.pendingError = err
        
class TimeStamp(TimeTicks, rfc2579.TextualConvention): pass

class TimeInterval(Integer, rfc2579.TextualConvention):
    subtypeSpec = Integer.subtypeSpec+ValueRangeConstraint(0, 2147483647)

class DateAndTime(rfc2579.TextualConvention, OctetString):
    subtypeSpec = OctetString.subtypeSpec+ValueSizeConstraint(8, 11)
    displayHint = "2d-1d-1d,1d:1d:1d.1d,1a1d:1d"

class StorageType(Integer, rfc2579.TextualConvention):
    subtypeSpec = Integer.subtypeSpec+SingleValueConstraint(1, 2, 3, 4, 5)
    namedValues = NamedValues(
        ('other', 1), ('volatile', 2), ('nonVolatile', 3),
        ('permanent', 4), ('readOnly', 5)
        )

class TDomain(ObjectIdentifier, rfc2579.TextualConvention): pass

class TAddress(OctetString, rfc2579.TextualConvention):
    subtypeSpec = OctetString.subtypeSpec+ValueSizeConstraint(1, 255)

mibBuilder.exportSymbols(
    'SNMPv2-TC',
    TextualConvention=rfc2579.TextualConvention, DisplayString=DisplayString,
    PhysAddress=PhysAddress, MacAddress=MacAddress, TruthValue=TruthValue,
    TestAndIncr=TestAndIncr, AutonomousType=AutonomousType,
    InstancePointer=InstancePointer, VariablePointer=VariablePointer,
    RowPointer=RowPointer, RowStatus=RowStatus, TimeStamp=TimeStamp,
    TimeInterval=TimeInterval, DateAndTime=DateAndTime,
    StorageType=StorageType, TDomain=TDomain, TAddress=TAddress
    )
